"""
Deal with the configuration of sync files and api keys etc
"""
import os

CONFIGURATION_DIRECTORIES = ['~/.config/gdrive_sync', '~/gdrive_sync']
CONFIGURATION_FILE = 'configuration.ini'
CREDENTIALS_FILE = 'credentials'
CHANGES_DB = 'gdrive_changes.db'

# According to Google it is OK for installed applications to distribute the client secret 
# in the source code, see: https://developers.google.com/accounts/docs/OAuth2InstalledApp
# If you are reading this and considering using the below code for another program, then
# please create your own access code, it takes only about 2 minutes, see: 
# https://developers.google.com/drive/quickstart-python for information about how
CLIENT_ID = '848223844858.apps.googleusercontent.com'
CLIENT_SECRET = '6Drtud3KlYscfVhrAxZzyamY'
REDIRECT_URI = 'urn:ietf:wg:oauth:2.0:oob'

class Configuration(object):
    def __init__(self, configuration_directory=None):
        if configuration_directory is None:
            configuration_directory = self.get_default_configuration_directory()
        self.configuration_directory = configuration_directory
        
        self.credentials_file = os.path.join(self.configuration_directory, CREDENTIALS_FILE)
        self.changes_file = os.path.join(self.configuration_directory, CHANGES_DB)
    
    def get_default_configuration_directory(self):
        # Check if any of the directories exist
        for path in CONFIGURATION_DIRECTORIES:
            pth = os.path.expanduser(path)
            if os.path.isdir(pth):
                return pth
        
        # No directory is present. Search for parent directories
        # and make our configuration directory in the first one
        # that exists
        for path in CONFIGURATION_DIRECTORIES:
            pth = os.path.expanduser(path)
            try:
                parent_pth = os.path.split(pth)[0]
                if os.path.isdir(parent_pth):
                    os.mkdir(pth)
                    return pth
            except OSError:
                continue
        
        raise OSError('Did not find configuration directory')



import os, datetime, re
from tzlocal import get_localzone
from collections import OrderedDict

from .configuration import Configuration
from .change_db import ChangeDB, GDriveFile, GDriveDirectory
from . import gdrive_communication as api

class GDriveSyncError(Exception):
    pass

SYSTEM_TZ = get_localzone()

class GDriveSyncDifference(object):
    ONLY_LOCAL = 'only_local'
    ONLY_REMOTE = 'only_remote'
    NEWER_LOCAL = 'newer_local'
    FILES_ARE_DIFFERENT = 'files_differ'
    FILE_DIRECTORY_MISMATCH = 'type_mismatch'

    def __init__(self, reason, remote_item=None, local_item=None):
        self.reason = reason
        self.remote_item = remote_item
        self.local_item = local_item

    def __str__(self):
        return "GDriveSyncDiff, Reason: %s, Local: %s, Remote: %s" % (self.reason, self.local_item, self.remote_item)
      
class LocalItem(object):
    def __init__(self, path):
        assert os.path.exists(path)
        self.path = path
        self._modified = None
        self._file_size = None
        self._is_file = None

    def __str__(self):
        return self.path

    @property
    def is_file(self):
        if self._is_file is None:
            self._is_file = os.path.isfile(self.path)
        return self._is_file

    @property
    def is_dir(self):
        return not self.is_file

    @property
    def modified(self):
        if self._modified is None:
            t = os.path.getmtime(self.path)
            self._modified = datetime.datetime.fromtimestamp(t, SYSTEM_TZ)
        return self._modified
    
    @property
    def file_size(self):
        if self._file_size is None:
            self._file_size = os.path.getsize(self.path)
        return self._file_size
  
class GDriveSync():
    def __init__(self, configuration_directory=None, verbose=False):
        self.config = Configuration(configuration_directory)
        self.change_db = ChangeDB(self.config.changes_file, logger=self.log)
        self.verbose = verbose
        self.root = None
        self.drive = None
        api.log = self.log
        self.logged_warnings = []
        self.logged_errors = []
        
    def connect(self):
        """
        Connect and authenticate with the Google Drive web service 
        """
        if self.drive is None:
            self.drive = api.connect(credentials_file=self.config.credentials_file)

    def is_connected(self):
        return self.drive is not None

    def update(self):
        """
        Update the local change database with any changes made to the Drive since last update
        """
        api.update(self.drive, self.change_db)
        
    def build_file_tree(self, root_id=None):
        """
        Build the google Drive file tree from the change list
        """
        if root_id is None:
            if not self.is_connected():
                self.connect()

            root_id = api.get_drive_root_id(self.drive)

        self.root = self.change_db.get_file_tree(root_id=root_id)
        
    def add_item(self, diff, local_base_dir, remote_base_dir, dry_run=False):
        """
        Add a file to google drive
        """
        # Get name on drive
        rel_path = os.path.relpath(os.path.abspath(diff.local_item.path),
                                   os.path.abspath(local_base_dir))
        rel_path = rel_path.replace(os.pathsep, '/')
        if rel_path == '.':
            drive_path = remote_base_dir
        else:
            drive_path = remote_base_dir + '/' + rel_path

        print drive_path

        drive_dir = os.path.split(drive_path)[0]
        print 'looking for', drive_dir
        try:
            parent = self.root.get_child(drive_dir)
        except KeyError:
            if dry_run:
                # Can't find the directory, if it's parent would also be created in this run.
                self.log('Would create %s in %s' % (drive_path, drive_dir))
                return
            else:
                raise Exception("Unknown behaviour")

        print 'parent is', parent.__str__()

        self.log('Creating  %s in %s' % (drive_path, drive_dir))
        if dry_run:
            return

        try:
            info = api.upload(self.drive, diff, parent.id)
        except api.errors.HttpError as e:
            self.log('Cannot add item %s' % diff.local_item, 'warning')
            self.log(e.message)
            return
        if diff.local_item.is_file:
            newitem = GDriveFile(info, parents=[parent])
        else:
            newitem = GDriveDirectory(info, parents=[parent])
        parent.add_child(newitem)

        #print info

    def update_item(self, diff, dry_run=False):
        """
        Upload a file (optionally update file that is already there)
        """

        assert diff.local_item.is_file

        self.log('Updating file %s from %s' % (diff.remote_item, diff.local_item))
        if dry_run:
            return

        try:
            _info = api.upload(self.drive, diff)
        except api.errors.HttpError as e:
            self.log('Cannot update file %s' % diff.local_item, 'warning')
            self.log(e.message)

    def delete_item(self, diff, remote_file_path, keep_multiple=False, dry_run=False):
        """
        Delete a file (move to trash)
        """
        # Get hold of 
        item = diff.remote_item
        if keep_multiple and len(item.parents) > 1:
            to_remove_parents = []
            for parent in item.parents:
                for path in parent.path:
                    if path.startswith('/' + remote_file_path):
                        to_remove_parents.append(parent)
            
            # remove duplicates
            to_remove_parents = list(OrderedDict.fromkeys(to_remove_parents))
            assert len(to_remove_parents) == 1
            to_remove_parent = to_remove_parents[0]
            try:
                api.delete(self.drive, diff, parent=to_remove_parent.id)
                to_remove_parent.unlink_child(item)
            except api.errors.HttpError as e:
                self.log('Cannot delete %s' % item, 'warning')
                self.log(str(e))
                return


        self.log('Deleting file %s' % item)
        
        if dry_run:
            return
        
        try:
            _info = api.delete(self.drive, diff)
        except api.errors.HttpError as e:
            self.log('Cannot delete %s' % item, 'warning')
            self.log(str(e))
            return
        
        for p in item.parents:
            p.unlink_child(item)

    def walk_drive(self, drive_dir):
        drive_dir_obj = self.root.get_child(drive_dir)

        if not drive_dir_obj.is_dir:
            raise GDriveSyncError('Requested path %r is not a directory on the Drive' % drive_dir)

        stack = [(child, [child.name]) for child in drive_dir_obj.children.itervalues()]
        while stack:
            item, pth = stack.pop()
            yield (item, pth)
            if item.is_dir:
                stack.extend((child, pth + [child.name]) for child in item.children.itervalues())

    def walk_syncable(self, drive_dir):
        for item, pth in self.walk_drive(drive_dir):
            dpath = [path for path in item.path if path.startswith('/' + drive_dir)]

            if len(dpath) == 1:
                # Only one path, fine!
                yield item, pth
            else:
                raise GDriveSyncError('Could not sync %s because of multiple parents within %s' % (item, drive_dir))

    def get_differences(self, drive_dir, local_dir, skip_dot_files_and_dirs=False, exclude=None):
        """
        Get the difference between local and remote directories
        
        Returns a list of GDriveSyncDifference objects
        """
        self.log('\nGetting differences between')
        self.log(' local directory: %s' % local_dir)
        self.log(' drive directory: %s' % drive_dir)
        
        local_dir = os.path.expanduser(local_dir)
        if not os.path.isdir(local_dir):
            raise GDriveSyncError('Local directory %s is not found' % os.path.abspath(local_dir))
        
        # List of differences that we find
        differences = []
        
        # Cache of local files and folders of class LocalItem 
        local_data = {}
        
        if exclude is None:
            exclude = []
        
        if skip_dot_files_and_dirs:
            # Exlude dot at start or \. (dos) or  /. (unix) 
            exclude.append('(^|[\\/])\.')
        
        ################################################################################
        # Get the Google Drive directory 
        try:
            self.root.get_child(drive_dir)
        except KeyError:
            # Top level directory does not exist
            #  -> Mark top level directory as missing 
            differences.append(GDriveSyncDifference(GDriveSyncDifference.ONLY_LOCAL,
                                                    local_item=LocalItem(os.path.abspath(local_dir))))
            #  -> Mark everything under the top level directory as missing
            for root, dirs, files in os.walk(unicode(local_dir)):
                for item in files + dirs:
                    pth_loc = os.path.join(root, item) 
                    if self._should_exclude(pth_loc, exclude):
                        # Skip excluded paths
                        continue
                    differences.append(GDriveSyncDifference(GDriveSyncDifference.ONLY_LOCAL,
                                                            local_item=LocalItem(pth_loc)))
            return differences
        
        ################################################################################
        # Find changes between remote and local
        for item, pth in self.walk_syncable(drive_dir):
            pth_loc = os.path.join(local_dir, *pth)
            
            if self._should_exclude(pth_loc, exclude):
                # This matched an exclude pattern
                continue
            
            # Check if the local item is present
            if pth_loc in local_data:
                local = local_data[pth_loc]
            elif os.path.exists(pth_loc):
                local = local_data[pth_loc] = LocalItem(pth_loc)
            else:
                differences.append(GDriveSyncDifference(GDriveSyncDifference.ONLY_REMOTE, remote_item=item))
                continue
            local_data[pth_loc] = local
            
            # Check that local and remote are either both files or both directories 
            if item.is_dir != local.is_dir:
                differences.append(GDriveSyncDifference(GDriveSyncDifference.FILE_DIRECTORY_MISMATCH,
                                                        remote_item=item,
                                                        local_item=local))
                continue
            
            # If the item is a directory it is considered to be equal on remote and local
            if item.is_dir:
                continue
            # If the item has been modified locally after previous upload then files are different
            elif local.modified > item.modified:
                differences.append(GDriveSyncDifference(GDriveSyncDifference.NEWER_LOCAL,
                                                        remote_item=item,
                                                        local_item=local))
                continue
            # If the file sizes differ then the files are different (obviously)
            elif local.file_size != item.file_size:
                print 'Different', local.path, local.file_size, item.file_size
                differences.append(GDriveSyncDifference(GDriveSyncDifference.FILES_ARE_DIFFERENT,
                                                        remote_item=item,
                                                        local_item=local))
                continue
        
        ################################################################################
        # Find items on local disk that are not present on the remote
        for root, dirs, files in os.walk(unicode(local_dir)):
            for item in files + dirs:
                pth_loc = os.path.join(root, item)
                
                if self._should_exclude(pth_loc, exclude):
                    # This matched an exclude pattern
                    continue
                
                # Have we already dealt with this above? In that case it is should be OK
                if pth_loc in local_data:
                    continue

                item = LocalItem(pth_loc)
                if item.file_size == 0:
                    self.log("Can't sync %s, because it is empty" % item, 'warning')
                    continue

                # File or directory is not on Drive
                differences.append(GDriveSyncDifference(GDriveSyncDifference.ONLY_LOCAL, local_item=item))
        
        return differences
    
    def _should_exclude(self, path, exclude_patterns):
        """
        Should this path be excluded?
        """
        for pat in exclude_patterns:
            if re.search(pat, path) is not None:
                self.log('Excluding "%s" due to exclude pattern "%s"' % (path, pat))
                return True
        return False
    
    def log(self, message, flag='info'):
        """
        By default messages are printed to stdin if verbose=True
        """
        prefix = ''
        if flag == 'warning':
            prefix = 'WARNING: '
            self.logged_warnings.append(message)
        elif flag == 'error':
            prefix = 'ERROR: '
            self.logged_errors.append(message)
        if self.verbose:
            print prefix + message
    


from __future__ import division
import os, random, time, json, webbrowser, mimetypes

import httplib2
from apiclient import errors
from apiclient.http import MediaFileUpload
from apiclient.discovery import build
from oauth2client.client import OAuth2WebServerFlow
from oauth2client.file import Storage

from .configuration import CLIENT_ID, CLIENT_SECRET, REDIRECT_URI

REQUEST_RATE_ERRORS = {'rateLimitExceeded', 'userRateLimitExceeded'}
OAUTH_SCOPE = 'https://www.googleapis.com/auth/drive'
MIME_DIRECTORY = 'application/vnd.google-apps.folder'
MIME_FILE = 'application/octet-stream'

FILE_FIELDS = ('id,title,mimeType,createdDate,modifiedDate,modifiedByMeDate,'
               'parents/id,originalFilename,fileExtension,fileSize,labels/trashed')
CHANGE_LIST_FIELDS = 'nextPageToken,largestChangeId,items(id,deleted,file(%s),fileId)' % FILE_FIELDS

# Default log function does nothing
log = lambda msg: None

def connect(credentials_file, client_id=CLIENT_ID, client_secret=CLIENT_SECRET, oauth_scope=OAUTH_SCOPE, redirect_uri=REDIRECT_URI):
    """
    Connect to Google Drive. Will initiate a dialog with the user if 
    the application is not approved by the user yet
    """
    log('Logging in ...')
    
    ###########################
    # Get stored credential
    
    storage = Storage(credentials_file)
    credentials = storage.locked_get()
    
    if credentials is None:
        authenticate(credentials_file, client_id, client_secret, oauth_scope, redirect_uri)
        credentials = storage.locked_get()
        assert credentials is not None, 'User did not approve access to the drive'
    
    ###########################
    # Connect
    
    http = httplib2.Http()
    
    credentials = storage.locked_get()
    if credentials.access_token_expired:
        credentials.refresh(http)
    
    http = credentials.authorize(http)
    drive_service = build('drive', 'v2', http=http)
    
    storage.locked_put(credentials)
    
    log('DONE\n')
    return drive_service

def authenticate(credentials_file, client_id, client_secret, oauth_scope=OAUTH_SCOPE, redirect_uri=REDIRECT_URI):
    flow = OAuth2WebServerFlow(client_id, client_secret, oauth_scope, redirect_uri, access_type='offline')
    authorize_url = flow.step1_get_authorize_url()
    print 'I will open your browser where you must accept that the program'
    print 'gets access to your Google Drive. It will give you a code that'
    print 'you must input below'
    
    webbrowser.open(authorize_url, 2)
    
    code = raw_input('Enter verification code: ').strip()
    credentials = flow.step2_exchange(code)
    
    # Check that the credentials work
    http = httplib2.Http()
    http = credentials.authorize(http)
    
    # Store the credentials
    storage = Storage(credentials_file)
    storage.locked_put(credentials)
    
def update(drive, database, max_changes_per_query=1000):
    """
    Update the change set database with any new changes on the Drive
    """
    
    def query_google_drive(request):
        t1 = time.time()
        results = call_with_exponential_backoff(request)
        items = results['items']
        
        if items:
            last = int(items[-1]['id'])
            largest = int(results['largestChangeId'])
            pst_done = (1 - (largest-last)/(largest-max_change_start)) * 100
            
            log(' Got %4d change items in %5.1f seconds (~%5.1f %% done)' % (len(items), time.time()-t1, pst_done))
        else:
            log(' No new changes found')
        
        return results, items
    
    log('Updating database ...')
    with database as db:
        max_change_start = db.max_change_id
        
        # Get new changes
        log(' Starting from change id %d' % (max_change_start+1))
        
        
        results = None
        num_downloaded = 0
        while True:
            if results is None:
                request = drive.changes().list(includeSubscribed=False,
                                               includeDeleted=True,
                                               pageToken=None,
                                               maxResults=max_changes_per_query,
                                               startChangeId=max_change_start+1,
                                               fields=CHANGE_LIST_FIELDS)
            else:
                request = drive.changes().list_next(request, results)
            
            # Are we done yet?
            if request is None:
                break
            
            # Download changes
            results, items = query_google_drive(request)
        
            # Store changes to database
            for change in items:
                db.add_change(change)
                num_downloaded += 1
    
    max_change_after = db.max_change_id    
    log('DONE: got %d changes, ended at change id %d\n' % (num_downloaded, max_change_after))

def upload(drive, diff, parent_id=None):
    """
    Upload a file to a folder with a given ID. 
    
    If is_direcory is True then a directory is created and the file
    name must be the name of the directory
    """
    update = not (diff.remote_item is None)
    assert update ^ (parent_id is not None)

    assert diff.local_item.is_dir ^ diff.local_item.file_size > 0

    body = {}
    if update and diff.local_item.is_file and diff.remote_item.is_file:
        mime = get_mime(diff.local_item.path)
        media_body = MediaFileUpload(diff.local_item.path, mimetype=mime, resumable=True)
        request = drive.files().update(fileId=diff.remote_item.id, body=body, media_body=media_body)

    elif not update:
        name = os.path.split(diff.local_item.path)[1]
        body = {'title': name, 'originalFilename': name, 'parents': [{'id': parent_id}]}

        if diff.local_item.is_file:
            mime = get_mime(diff.local_item.path)
            media_body = MediaFileUpload(diff.local_item.path, mimetype=mime, resumable=True)
        else:
            assert diff.local_item.is_dir
            media_body = None
            body['mimeType'] = MIME_DIRECTORY

        request = drive.files().insert(body=body, media_body=media_body)
    else:
        # update and diff.remote_item.is_dir                             -- A directory can't be updated
        # update and diff.remote_item.is_dir and diff.local_item.is_file -- Special case of the prev., should never happen
        raise NotImplementedError("Not implemented case")

    drive_file_info = call_with_exponential_backoff(request)
    return drive_file_info

def delete(drive, diff, parent=None, trash=True):
    """
    Either move a file to trash, permanently delete it, or remove one parent from it
    """
    if trash and parent is None:
        request = drive.files().trash(fileId=diff.remote_item.id)
    elif not trash and parent is None:
        request = drive.files().delete(fileId=diff.remote_item.id)
    elif parent is not None:
        request = drive.parents().delete(fileId=diff.remote_item.id, parentId=parent)
    return call_with_exponential_backoff(request)

def call_with_exponential_backoff(action):
    """
    Execute with exponential back-off
    Adapted from Googles API documentation
    """
    N = 5
    for n in range(N):
        try:
            return action.execute()
        except errors.HttpError as e:
            error = json.loads(e.content)
            # Get error code and type
            code = error.get('code')
            reason = error.get('error').get('errors')[0].get('reason')
            # Is this rate limiting errors?
            if n < N-1 and code == 403 and reason in REQUEST_RATE_ERRORS:
                # Apply exponential back-off.
                secs = (2 ** n) + random.randint(0, 1000) / 1000
                log('Google is limiting our query rate, backing off for %.2f seconds' % secs)
                time.sleep(secs)
            else:
                # Other error or end of tries, re-raise.
                raise


def get_drive_root_id(drive):
    action = drive.about().get()
    r = call_with_exponential_backoff(action)
    return r['rootFolderId']


def get_mime(filename):
    type, encoding = mimetypes.guess_type(filename, strict=False)
    return type if type is not None else MIME_FILE



import os
import cPickle as pickle
import dateutil.parser
from collections import defaultdict

from . import gdrive_communication as api

class ChangeDB(object):
    def __init__(self, filename, logger=None):
        """
        This class stores the Google Drive file changes. The format of the
        changes is identical with the dictionaries returned by the Google
        Python API code for Google Drive
        """
        self.filename = filename
        if logger is None:
            self.logger = lambda msg, flag=None: None
        else:
            self.logger = logger

        self.database = None
        
    @property
    def changes(self):
        return self.database['changes']
    
    @property
    def max_change_id(self):
        return self.database['max_change_id']
    
    def add_change(self, change):
        """
        Add change to the database
        """
        cid = int(change['id'])
        assert cid not in self.database['changes']
        assert cid > self.database['max_change_id']
         
        self.database['max_change_id'] = cid
        self.database['changes'][cid] = change
    
    def __read(self):
        """
        Load change database from disk
        """
        if os.path.isfile(self.filename):
            # Read database from file
            with open(self.filename, 'rb') as f:
                self.database = pickle.load(f)
        else:
            self.database = {'max_change_id': 0, 'changes': {}}
            self.logger('Creating new change set database')
            # Check that we can write to the database file
            self.__write()
        
        self.logger('Loaded changes database containing %d changes. Last change id is %d' % (len(self.changes), self.max_change_id))
    
    def get_roots(self):
        """
        Replay the stored changes in the database to create a local
        representation of the current state of the remote Drive
        """
        if self.database is None:
            self.__read()

        changes = self.changes
        id_to_obj_map = {}
        deleted = set()
        if not changes:
            return GDriveDirectory(None, None)
        
        keys = sorted(changes.keys())
        sorted_changes = [self.changes[cid] for cid in keys]
        
        roots = []
        
        for c in sorted_changes:
            fid = c['fileId']
            if c.get('deleted', False):
                if fid in id_to_obj_map:
                    obj = id_to_obj_map[fid]
                    obj.delete()
                    del id_to_obj_map[fid]
                else:
                    pass
                    #print 'Object to delete is not in map', c
                deleted.add(fid)
                continue
                
            f = c['file']
            assert f['id'] == fid

            parents = list()

            for parent in f['parents']:
                parent_id = parent['id']

                if parent_id in id_to_obj_map:
                    parents.append(id_to_obj_map[parent_id])
                else:
                    p_temp = GDriveDirectory({'id': parent_id}, [])
                    parents.append(p_temp)
                    id_to_obj_map[parent_id] = p_temp
                    #print 'Found unrooted object %s' % f['title'], fid
                    roots.append(p_temp)
            
            if fid in id_to_obj_map:
                obj = id_to_obj_map[fid]
                if obj.is_root:
                    #print 'Rooting object %s' % f['title'], fid
                    roots.remove(obj)
                obj.update(f, parents)
            else:
                if f['mimeType'] == api.MIME_DIRECTORY:
                    obj = GDriveDirectory(f, parents)
                else:
                    obj = GDriveFile(f, parents)

                for p in parents:
                    p.add_child(obj)

                id_to_obj_map[fid] = obj
        
        roots = [s for s in roots if s.id not in deleted]

        return roots

    def get_file_tree(self, root_id=None, include_trashed=False):
        roots = self.get_roots()
        root = None
        for r in roots:
            if r.id == root_id:
                root = r
                break

        assert root.is_root
        assert len(root.parents) == 0

        # Remove trashed before verify - it might be duplicates where one is trashed and the other not
        if not include_trashed:
            root.remove_all_trashed()
        
        # Check that everything is OK with the constructed file tree
        root.verify(self.logger)
        
        return root
    
    def __write(self):
        """
        Save change database to disk
        """
        with open(self.filename, 'wb') as f:
            pickle.dump(self.database, f, protocol=pickle.HIGHEST_PROTOCOL)
        self.logger('Saved changes database containing %d changes. Last change id is %d' % (len(self.changes), self.max_change_id))
    
    def __enter__(self):
        """
        Allow the database to be used as a context manager
        """
        self.__read()
        return self
    
    def __exit__(self, *argv):
        """
        Allow the database to be used as a context manager
        make sure that the database is saved to disk
        """
        self.__write()
        if argv[0] is not None:
            self.logger('Emergency saved changes database due to exception')

class GDriveFile(object):
    is_file = True
    is_dir = False
    is_deleted = False
    
    def __init__(self, info, parents=None):
        self.parents = [] if parents is None else parents
        self.info = info
        self.is_root = len(self.parents) == 0
    
    @property
    def id(self):
        return self.info['id']

    def __str__(self):
        p = self.path
        if len(p) == 1:
            return p[0]
        else:
            return str(p)

    @property
    def name(self):
        if self.is_root:
            return '/'
        
        orig = self.info.get('originalFilename', None)
        title = self.info.get('title', None)
        ext = self.info.get('fileExtension', None)
        
        if orig:
            return orig
        elif ext and title.endswith(ext):
            return title
        elif ext:
            return '%s.%s' % (title, ext)
        else:
            return title
    
    @property
    def path(self):
        "The full path to this drive item"
        if self.is_root:
            return ['']

        paths = []
        for parent in self.parents:
            for p in parent.path:
                paths.append(p + "/" + self.name)

        return paths

    @property
    def modified(self):
        isotime = self.info['modifiedDate'] # Format: 2013-01-16T15:07:34.915Z
        return dateutil.parser.parse(isotime)
    
    @property
    def file_size(self):
        size = self.info['fileSize']
        return int(size)
    
    @property
    def is_trashed(self):
        return self.info['labels']['trashed']
    
    def delete(self):
        if self.parents is not None:
            for p in self.parents:
                p.unlink_child(self)

        self.is_deleted = True
    
    def update(self, new_info, new_parents):
        # Swap parents if necessary
        assert len(new_info['parents']) == len(new_parents)

        new_parent_ids = [p['id'] for p in new_info['parents']]
        for new_parent_id in new_parent_ids:
            assert new_parent_id in [p.id for p in new_parents]

        current_parent_ids = []
        if self.is_root:
            assert 'parents' not in self.info
            assert len(self.parents) == 0
        else:
            # check whether the parent information is consistent
            current_parent_ids = set([p['id'] for p in self.info['parents']])
            self_parent_ids = set([p.id for p in self.parents])
            assert len(current_parent_ids.symmetric_difference(self_parent_ids)) == 0

        if set(new_parent_ids) != set(current_parent_ids):
            self._swap_parents(new_parents)
        
        # Update information
        self.is_root = False # root object will never be updated
        self.info = new_info
    
    def _swap_parents(self, new_parents):
        for p in self.parents:
            p.unlink_child(self)

        if new_parents is not None:
            for p in new_parents:
                p.add_child(self)

            self.is_root = False
        else:
            self.is_root = True

        self.parents = new_parents
    
    def verify(self, logger):
        assert self.info is not None
        assert not self.is_deleted
        if self.is_root:
            assert self.is_dir and not self.is_file
            assert len(self.parents) == 0
        else:
            assert len(self.parents) > 0
            assert all([self.id in p.children for p in self.parents])

            for parent in self.parents:

                if not parent.is_root:
                    assert parent.id in [p['id'] for p in self.info['parents']]
        
            if self.is_dir:
                assert not self.is_file
                assert self.info['mimeType'] == api.MIME_DIRECTORY
            else:
                assert self.is_file
                assert self.info['mimeType'] != api.MIME_DIRECTORY

class GDriveDirectory(GDriveFile):
    is_file = False
    is_dir = True
    
    def __init__(self, info, parents):
        super(GDriveDirectory, self).__init__(info, parents)
        self.children = {}
    
    def get_children(self):
        return self.children.values()            
            
    def add_child(self, child):
        self.children[child.info['id']] = child
    
    def unlink_child(self, child):
        del self.children[child.info['id']]
    
    def get_child(self, path):
        """
        Get a child or a grand child, great grand child etc. The path
        is interpreted in the Unix manner with separator '/'
        """
        if '/' in path:
            # This is a path to a sub-directory
            pth_elems = path.split('/')
            if pth_elems[0] == '':
                assert self.is_root
                pth_elems = pth_elems[1:]
            
            elem = self
            for name in pth_elems:
                if not elem.is_dir:
                    raise KeyError('Item "%s" is a file, not a directory' % elem.path)
                if name == '':
                    continue
                elem = elem[name]
            return elem
        else:
            # This is a name to a item in this directory
            for c in self.children.itervalues():
                if c.name == path:
                    return c
            raise KeyError('Item "%s" in "%s" not found on Drive' % (path, self.path))
    __getitem__ = get_child
    
    def has_file(self, filename):
        for c in self.children.itervalues():
            if c.name == filename:
                return c.is_file
        return False
    
    def has_dir(self, filename):
        for c in self.children.itervalues():
            if c.name == filename:
                return c.is_dir
        return False
    
    def remove_all_trashed(self):
        stack = self.children.values()
        while stack:
            item = stack.pop()
            if item.is_trashed:
                item.delete()
            if item.is_dir:
                stack.extend(item.children.itervalues())
    
    def verify(self, logger):
        super(GDriveDirectory, self).verify(logger)        
        names = defaultdict(list)
        for cid, c in self.children.iteritems():
            assert cid == c.info['id']
            names[c.name].append(c)
            
            if self not in c.parents:
                print 'Item "%s" claims "%s" as its parents, but is found in "%s"' % (c.name, c.parent.path, self.path)
            
            c.verify(logger)
        
        for name, children in names.iteritems():
            if len(children) == 1:
                continue
            logger('There are %d items called "%s" on Drive in "%s"' % (len(children), name, self.path), 'warning')



"""
Command line module for gdrive_sync
"""
from __future__ import absolute_import
import sys, argparse, time, logging
from gdrive_sync import GDriveSync, GDriveSyncError, GDriveSyncDifference as Difference

hdlr = logging.StreamHandler()
hdlr.setLevel(logging.WARNING)
logging.getLogger().addHandler(hdlr)

def update(args, gds=None):
    """
    Update the database of changes with whatever has changed since the last update
    """
    if gds is None:
        gds = GDriveSync(verbose=True)

    gds.connect()
    gds.update()
    
    log_problems(gds)

def print_differences(args, gds=None):
    """
    Print changes between local directory and what is on the Drive
    """
    t1 = time.time()
    print '\n----------------- RUNNING -------------------'
    if gds is None:
        gds = GDriveSync(verbose=True)
    try:
        gds.build_file_tree()
        differences = gds.get_differences(args.remote_dir, args.local_dir, 
                                          exclude=args.exclude)
    except GDriveSyncError as e:
        print '\nERROR:\n    %s' % e.message
        sys.exit(1)
    print 'DONE in %.1f seconds' % (time.time() - t1)
    print '-------------- END OF RUNNING -----------------'
    
    print '\n---------------- DIFFERENCES ------------------'
    for diff in differences:
        if diff.reason == Difference.ONLY_LOCAL:
            print 'Only on local disk: %s' % diff.local_item
        elif diff.reason == Difference.ONLY_REMOTE:
            print 'Only on Google Drive: %s' % diff.remote_item
        elif diff.reason == Difference.NEWER_LOCAL:
            print 'Newer on local disk: %s' % diff.local_item
        elif diff.reason == Difference.FILES_ARE_DIFFERENT:
            print 'Files are different: %s, %s' % (diff.remote_item, diff.local_item)
        else:
            print '\nERROR:\n    Unknown difference type %s' % diff.reason
            sys.exit(1)
    print '-------------- END OF DIFFERENCES --------------'
    
    log_problems(gds)
    
    return differences

def sync_up(args):
    """
    Upload missing files and directories to google Drive
    """
    gds = GDriveSync(verbose=True)
    # First, update the database
    update(args, gds=gds)
    
    # Then, print the differences between local and remote directory
    differences = print_differences(args, gds=gds)
    
    allow_delete = args.delete
    
    if not differences:
        print 'No differences found!'
        return
    
    print '\n-------------------- ASK ----------------------'
    print '\nDo you want to synchronize from local disk to Google Drive?'
    print 'Local directory:       ', args.local_dir
    print 'Google Drive directory:', args.remote_dir
    if allow_delete:
        print 'The program will delete files on Google Drive that are not present locally!'
    else:
        print 'The program will NOT delete files on Google Drive that are not present locally'
    if not (args.yes or args.dry_run):
        yN = raw_input('Continue [y/N]? ')
        if yN.strip().lower() not in ('y', 'yes'):
            print 'Will do nothing'
            return
    
    print '\n-------------------- SYNC ---------------------'

    for diff in differences:
        print
        if diff.reason == Difference.ONLY_LOCAL:
            gds.add_item(diff, args.local_dir, args.remote_dir, dry_run=args.dry_run)
        elif diff.reason == Difference.ONLY_REMOTE:
            if allow_delete:
                gds.delete_item(diff, args.remote_dir, keep_multiple=args.keep_multiple, dry_run=args.dry_run)
        elif diff.reason == Difference.NEWER_LOCAL:
            gds.update_item(diff, dry_run=args.dry_run)
        elif diff.reason == Difference.FILES_ARE_DIFFERENT:
            print 'Files are different: %s, %s (doing nothing)' % (diff.remote_item, diff.local_item)
        else:
            print '\nERROR:\n    Unknown difference type %s' % diff.reason
            sys.exit(1)
    
    print '-------------------- DONE ---------------------'
    log_problems(gds)
    
def log_problems(gds):
    print '\n------------------- STATUS ---------------------'
    print 'Finished with %d warnings and %d errors' % (len(gds.logged_warnings), len(gds.logged_errors))
    for msg in gds.logged_warnings:
        print 'WARNING: %s' % msg
    for msg in gds.logged_errors:
        print 'ERROR: %s' % msg
    print '---------------- END OF STATUS ----------------'

def command_line_interface():
    # create the top-level parser
    parser = argparse.ArgumentParser(description='Synchronize directories with Google Drive')
    subparsers = parser.add_subparsers()
    
    parser_update = subparsers.add_parser('update')
    parser_update.set_defaults(func=update)
    
    parser_diff = subparsers.add_parser('diff')
    parser_diff.set_defaults(func=print_differences)
    parser_diff.add_argument('remote_dir', default='/', nargs='?', help='The directory on Google Drive')
    parser_diff.add_argument('local_dir', default='.', nargs='?', help='The local directory')
    
    parser_sync_up = subparsers.add_parser('sync-up')
    parser_sync_up.set_defaults(func=sync_up)
    parser_sync_up.add_argument('remote_dir', default='/', nargs='?', help='The directory on Google Drive')
    parser_sync_up.add_argument('local_dir', default='.', nargs='?', help='The local directory')
    parser_sync_up.add_argument('-d', '--dry-run', action='store_true', dest='dry_run',
                                help='print what will be done, but do not do it')
    parser_sync_up.add_argument('-y', action='store_true', dest='yes',
                                help='do not ask if you want to sync, just do it')
    parser_sync_up.add_argument('-x', '--delete', action='store_true', dest='delete',
                                help='allow deleting files (default off)')
    parser_sync_up.add_argument('-k', '--keep-other-parents', action='store_true', dest='keep_multiple',
                                help='If files or directories are deleted (-x or --delete) and have '
                               'multiple parent folders: Should they be delete from only one place or all (default)?')
    
    for subparser in (parser_diff, parser_sync_up):
        subparser.add_argument('-e', '--exclude', action='append', dest='exclude', default=[],
                               help='regex pattern for file names to exclude')
    
    # Parse the arguments and call whatever function was selected
    args = parser.parse_args()
    args.func(args)

if __name__ == '__main__':
    command_line_interface()



from setuptools import setup

with open('README.rst') as f:
    long_description = f.read()

setup(
    name = 'gdrive_sync',
    packages = ['gdrive_sync'],
    version = '0.6.1',
    description = 'Sync directories between your computer and Google Drive',
    long_description = long_description,
    author = 'Tormod Landet',
    url = 'https://bitbucket.org/trlandet/gdrive_sync',
    classifiers = ['Development Status :: 4 - Beta',
                   'Intended Audience :: End Users/Desktop',
                   'License :: OSI Approved :: BSD License',
                   'License :: OSI Approved :: GNU General Public License (GPL)',
                   'License :: OSI Approved :: Python Software Foundation License',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python :: 2.7',
                   'Topic :: System :: Archiving :: Mirroring',
                   'Topic :: Software Development :: Libraries :: Python Modules',
                   'Environment :: Console'],
    install_requires = ['google-api-python-client', 'python-dateutil', 'tzlocal'],
    scripts = ['scripts/gdrive_sync']
)

